#include <stdlib.h>
#include <stdio.h>
#include <getopt.h>
#include <errno.h>
#include <string.h>
#include "heat.h"

/* 
 * Taken in the number of command line arguments and a pointer to strings
 * containing the actual arguments. Returns a userArgs_t struct containing
 * the parsed arguments. Returns 0 for success, -1 for failure.
 */
int get_user_args(
        const int argc,
        char **argv,
        userArgs_t *userArgs
        ) {
    int option;

    /* Parameters (what the model is given) */
    userArgs->timeStepCount      = 3;
    userArgs->rowCount           = 3;
    userArgs->columnCount        = 3;
    userArgs->initialTemperature = 100.0;

    /* get user input */
    while ((option = getopt(argc, argv, "t:r:c:i:")) != -1) {
        switch(option) {
            case 't':
                errno = 0;
                userArgs->timeStepCount = strtol(optarg, NULL, 10);
                if (errno) {
                    fprintf(stderr, "%s: timeStepCount conversion failed: %s\n",
                            argv[0], strerror(errno));
                    return -1;
                }
                break;
            case 'r':
                errno = 0;
                userArgs->rowCount = strtol(optarg, NULL, 10);
                if (errno) {
                    fprintf(stderr, "%s: rowCount conversion failed: %s\n",
                            argv[0], strerror(errno));
                    return -1;
                }
                break;
            case 'c':
                errno = 0;
                userArgs->columnCount = strtol(optarg, NULL, 10);
                if (errno) {
                    fprintf(stderr, "%s: columnCount conversion failed: %s\n",
                            argv[0], strerror(errno));
                    return -1;
                }
                break;
            case 'i':
                errno = 0;
                userArgs->initialTemperature = strtof(optarg, NULL);
                if (errno) {
                    fprintf(stderr,
                            "%s: initialTemperature conversion failed: %s\n",
                            argv[0], strerror(errno));
                    return -1;
                }
                break;
            default:
                fprintf(stderr, "Usage: %s -t <time-step-count> -r <row-count> -c    <column-count> -i <initial-temperature>\n",
                        argv[0]);
                return -1;
        }
    }

    /* valiate input */
    if (userArgs->timeStepCount <= 0) {
        fprintf(stderr, "%s: cannot simulate with < 1 time step\n", argv[0]);
        errno = EINVAL;
    }
    if (userArgs->rowCount <= 0) {
        fprintf(stderr, "%s: cannot simulate with < 1 row\n", argv[0]);
        errno = EINVAL;
    }
    if (userArgs->columnCount <= 0) {
        fprintf(stderr, "%s: cannot simulate with < 1 column\n", argv[0]);
        errno = EINVAL;
    }
    if (userArgs->initialTemperature < 0) {
        fprintf(stderr, "%s: cannot simulate with < 0 temperature\n", argv[0]);
        errno = EINVAL;
    }
    if (errno == EINVAL) {
        return -1;
    }

    /* 
     * 1: calculate number of points, store in shared struct, since
     * this will be used a lot.
     */
    userArgs->pointCount = userArgs->rowCount * userArgs->columnCount;

    /* Cache process name, makes for better error messages */
    userArgs->procName = argv[0];

    return 0;
}

/* 
 * Takes in userArgs_t struct, allocates memory for the field of points,
 * sets initial conditions, and returns the field of points as a point_t
 * struct. Returns NULL on error.
 */
point_t *init_points(const userArgs_t userArgs) {
    point_t *points;
    unsigned int pointIndex;

    /* allocate memory */
    if((points = (point_t*)malloc(userArgs.pointCount * sizeof(point_t)))
            == NULL) {
        fprintf(stderr, "%s: Failed too allocate %zu bytes for grid: %s\n",
                userArgs.procName,
                (userArgs.pointCount * sizeof(point_t)),
                strerror(errno)
               );
        return NULL;
    }

    /* Count the number of neighbors, since we will be using that number
     * a lot
     */
#ifdef OPENMP
#pragma omp parallel for 
#endif
    for (
            /* 2: start at point 0 */
            pointIndex = 0;

            /* 3: is point outside the object? */
            pointIndex < userArgs.pointCount;

            /* 10: move on to the next point */
            pointIndex++
        ) {

        /* 4a: Left column gets the hot, initial temperature */
        if (pointIndex % userArgs.columnCount == 0) {
            points[pointIndex].temperatureNow = userArgs.initialTemperature;
        }
        /* 4b: Everyone else starts at 0 */
        else {
            points[pointIndex].temperatureNow = 0.0;
        }

        /* 5: assume 4 neighbors */
        points[pointIndex].neighborCount = 4;

        /* 6: top row has 1 less neighbor */
        if (pointIndex < userArgs.columnCount) {
            points[pointIndex].neighborCount--;
        }

        /* 7: bottom row has 1 less neighbor */
        if (pointIndex >= (userArgs.rowCount-1) * userArgs.columnCount) {
            points[pointIndex].neighborCount--;
        }

        /* 8: left column has 1 less neighbor */
        if (pointIndex % userArgs.columnCount == 0) {
            points[pointIndex].neighborCount--;
        }

        /* 9: right column has 1 less neighbor */
        if (pointIndex % userArgs.columnCount == userArgs.columnCount-1) {
            points[pointIndex].neighborCount--;
        }
    }

    return points;
}

void print_points(const userArgs_t userArgs,const point_t *points) {
    unsigned int pointIndex;

    /* Don't parallelize this to keep print order consistent */
    for (
            /* 13: start at point 0 */
            pointIndex = 0;

            /* 14: is point outside the object? */
            pointIndex < userArgs.pointCount;

            /* 17: move on to the next point */
            pointIndex++
        ) {

        /* 15: print temperature */
        printf("%.2f ", points[pointIndex].temperatureNow);

        /* 16: print newline between rows */
        if (pointIndex % userArgs.columnCount == userArgs.columnCount-1) {
            printf("\n");
        }
    }


    /* 18: print newline between time steps */
    printf("\n");
}

void set_future_points(const userArgs_t userArgs,point_t *points) {
    unsigned int pointIndex;

    /* 
     * Set temperatures in the future grid based on temperatures in
     * the present grid
     */
#ifdef OPENMP
#pragma omp parallel for 
#endif
    for (
            /* 19: start at point 0 */
            pointIndex = 0;

            /* 20: is point outside the object? */
            pointIndex < userArgs.pointCount;

            /* 27: move on to the next point */
            pointIndex++
        ) {

        /* 21: initialize next temperature */
        points[pointIndex].temperatureNext = 0.0;

        /* 22: add top neighbor */
        if (pointIndex >= userArgs.columnCount) {
            points[pointIndex].temperatureNext +=
                points[pointIndex-userArgs.columnCount].temperatureNow;
        }

        /* 23: add bottom neighbor */
        if (pointIndex < (userArgs.rowCount-1) * userArgs.columnCount) {
            points[pointIndex].temperatureNext +=
                points[pointIndex+userArgs.columnCount].temperatureNow;
        }

        /* 24: add left neighbor */
        if (pointIndex % userArgs.columnCount > 0) {
            points[pointIndex].temperatureNext +=
                points[pointIndex-1].temperatureNow;
        }

        /* 25: add right neighbor */
        if (pointIndex % userArgs.columnCount < userArgs.columnCount-1) {
            points[pointIndex].temperatureNext +=
                points[pointIndex+1].temperatureNow;
        }

        /* 26: divide by number of neighbors */
        points[pointIndex].temperatureNext /=
            points[pointIndex].neighborCount;
    }
}

void copy_future_to_present(const userArgs_t userArgs,point_t *points) {
    unsigned int pointIndex;
#ifdef OPENMP
#pragma omp parallel for 
#endif
    for (
            /* 28: start at point 0 */
            pointIndex = 0;

            /* 29: is point outside the object? */
            pointIndex < userArgs.pointCount;

            /* 31: move on to the next point */
            pointIndex++
        ) {

        /* 30: advance temperature */
        points[pointIndex].temperatureNow =
            points[pointIndex].temperatureNext;
    }
}
